@@ -259,6 +259,40 @@ int config_volume_power = 2;
 int config_feed_samplerate = 48000;
 int config_min_buffer = 30000;
 
+#define EQ_BAND_COUNT 10
+#define EQ_AUTO_HEADROOM_MARGIN_DB 1.0f
+#define LIMITER_THRESHOLD 0.89125093813f // -1 dBFS
+#define LIMITER_ATTACK_MS 1.5f
+#define LIMITER_RELEASE_MS 120.0f
+
+typedef struct {
+	float b0;
+	float b1;
+	float b2;
+	float a1;
+	float a2;
+	float z1_l;
+	float z2_l;
+	float z1_r;
+	float z2_r;
+} eq_biquad_t;
+
+static const float eq_band_freqs[EQ_BAND_COUNT] = {
+	31.25f, 62.5f, 125.0f, 250.0f, 500.0f, 1000.0f, 2000.0f, 4000.0f, 8000.0f, 16000.0f
+};
+eq_biquad_t eq_bands[EQ_BAND_COUNT];
+float eq_band_gain_db[EQ_BAND_COUNT] = {0};
+int eq_enabled = 0;
+int eq_active = 0;
+int eq_coeff_sample_rate = 0;
+bool eq_dirty = true;
+float eq_headroom_db = 0.0f;
+float eq_headroom_gain = 1.0f;
+float limiter_gain = 1.0f;
+float limiter_attack_coeff = 0.0f;
+float limiter_release_coeff = 0.0f;
+int limiter_coeff_sample_rate = 0;
+
 unsigned int test1 = 0;
 
 enum status {
@@ -1345,6 +1379,209 @@ void stop_decoder() {
 	decoder_allocated = 0;
 }
 
+static void eq_reset_state() {
+	for (int i = 0; i < EQ_BAND_COUNT; i++) {
+		eq_bands[i].z1_l = 0.0f;
+		eq_bands[i].z2_l = 0.0f;
+		eq_bands[i].z1_r = 0.0f;
+		eq_bands[i].z2_r = 0.0f;
+	}
+}
+
+static void limiter_reset_state() {
+	limiter_gain = 1.0f;
+}
+
+static void limiter_update_coefficients(int sample_rate) {
+	if (sample_rate <= 0) return;
+
+	float attack_samples = (LIMITER_ATTACK_MS * 0.001f) * (float)sample_rate;
+	float release_samples = (LIMITER_RELEASE_MS * 0.001f) * (float)sample_rate;
+
+	if (attack_samples < 1.0f) attack_samples = 1.0f;
+	if (release_samples < 1.0f) release_samples = 1.0f;
+
+	limiter_attack_coeff = expf(-1.0f / attack_samples);
+	limiter_release_coeff = expf(-1.0f / release_samples);
+	limiter_coeff_sample_rate = sample_rate;
+}
+
+static float eq_biquad_magnitude(const eq_biquad_t *f, float w) {
+	float cos_w = cosf(w);
+	float sin_w = sinf(w);
+	float cos_2w = cosf(2.0f * w);
+	float sin_2w = sinf(2.0f * w);
+
+	float nr = f->b0 + (f->b1 * cos_w) + (f->b2 * cos_2w);
+	float ni = -(f->b1 * sin_w) - (f->b2 * sin_2w);
+	float dr = 1.0f + (f->a1 * cos_w) + (f->a2 * cos_2w);
+	float di = -(f->a1 * sin_w) - (f->a2 * sin_2w);
+
+	float den = (dr * dr) + (di * di);
+	if (den <= 1e-20f) return 1.0f;
+
+	float mag2 = ((nr * nr) + (ni * ni)) / den;
+	if (mag2 <= 0.0f || !isfinite(mag2)) return 1.0f;
+
+	return sqrtf(mag2);
+}
+
+static void eq_update_auto_headroom(int sample_rate) {
+	const float two_pi = 6.28318530717958647692f;
+	const float min_freq = 20.0f;
+	const int sweep_points = 192;
+
+	eq_headroom_db = 0.0f;
+	eq_headroom_gain = 1.0f;
+
+	if (!eq_enabled || !eq_active || sample_rate <= 0) return;
+
+	float max_freq = sample_rate * 0.49f;
+	if (max_freq <= min_freq) return;
+
+	float ratio = max_freq / min_freq;
+	float max_mag = 1.0f;
+
+	for (int i = 0; i < sweep_points; i++) {
+		float t = i / (float)(sweep_points - 1);
+		float freq = min_freq * powf(ratio, t);
+		float w = two_pi * (freq / (float)sample_rate);
+		float total_mag = 1.0f;
+
+		for (int band = 0; band < EQ_BAND_COUNT; band++) {
+			if (fabsf(eq_band_gain_db[band]) < 0.01f) continue;
+			total_mag *= eq_biquad_magnitude(&eq_bands[band], w);
+		}
+
+		if (isfinite(total_mag) && total_mag > max_mag) max_mag = total_mag;
+	}
+
+	if (max_mag > 1.0f) {
+		eq_headroom_db = (20.0f * log10f(max_mag)) + EQ_AUTO_HEADROOM_MARGIN_DB;
+		eq_headroom_gain = powf(10.0f, -eq_headroom_db / 20.0f);
+	}
+}
+
+static void eq_set_identity(eq_biquad_t *f) {
+	f->b0 = 1.0f;
+	f->b1 = 0.0f;
+	f->b2 = 0.0f;
+	f->a1 = 0.0f;
+	f->a2 = 0.0f;
+}
+
+static void eq_rebuild_coefficients(int sample_rate) {
+	const float q = 1.41421356237f;
+	const float two_pi = 6.28318530717958647692f;
+	bool reset_state = (eq_coeff_sample_rate != sample_rate);
+
+	if (sample_rate <= 0) return;
+
+	eq_active = 0;
+
+	for (int i = 0; i < EQ_BAND_COUNT; i++) {
+		float gain_db = eq_band_gain_db[i];
+		float freq = eq_band_freqs[i];
+		eq_biquad_t *f = &eq_bands[i];
+
+		if (gain_db > 12.0f) gain_db = 12.0f;
+		if (gain_db < -12.0f) gain_db = -12.0f;
+		eq_band_gain_db[i] = gain_db;
+
+		if (fabsf(gain_db) < 0.01f || freq >= (sample_rate * 0.49f)) {
+			eq_set_identity(f);
+			continue;
+		}
+
+		float w0 = two_pi * (freq / (float) sample_rate);
+		float cw = cosf(w0);
+		float sw = sinf(w0);
+		float alpha = sw / (2.0f * q);
+		float a = powf(10.0f, gain_db / 40.0f);
+		float b0 = 1.0f + (alpha * a);
+		float b1 = -2.0f * cw;
+		float b2 = 1.0f - (alpha * a);
+		float a0 = 1.0f + (alpha / a);
+		float a1 = -2.0f * cw;
+		float a2 = 1.0f - (alpha / a);
+
+		if (a0 == 0.0f || !isfinite(b0) || !isfinite(b1) || !isfinite(b2) || !isfinite(a1) || !isfinite(a2)) {
+			eq_set_identity(f);
+			continue;
+		}
+
+		f->b0 = b0 / a0;
+		f->b1 = b1 / a0;
+		f->b2 = b2 / a0;
+		f->a1 = a1 / a0;
+		f->a2 = a2 / a0;
+		eq_active = 1;
+	}
+
+	eq_coeff_sample_rate = sample_rate;
+	eq_dirty = false;
+	eq_update_auto_headroom(sample_rate);
+	if (reset_state) {
+		eq_reset_state();
+		limiter_reset_state();
+		limiter_update_coefficients(sample_rate);
+	}
+}
+
+static inline float eq_process_biquad(float x, eq_biquad_t *f, bool left) {
+	float *z1 = left ? &f->z1_l : &f->z1_r;
+	float *z2 = left ? &f->z2_l : &f->z2_r;
+	float y = (f->b0 * x) + *z1;
+	*z1 = (f->b1 * x) - (f->a1 * y) + *z2;
+	*z2 = (f->b2 * x) - (f->a2 * y);
+	return y;
+}
+
+static inline void eq_process_stereo(float *l, float *r) {
+	if (!eq_enabled || !eq_active) return;
+
+	float ll = *l;
+	float rr = *r;
+	for (int i = 0; i < EQ_BAND_COUNT; i++) {
+		ll = eq_process_biquad(ll, &eq_bands[i], true);
+		rr = eq_process_biquad(rr, &eq_bands[i], false);
+	}
+
+	*l = ll;
+	*r = rr;
+}
+
+static inline void limiter_process_stereo(float *l, float *r) {
+	if (!eq_enabled || !eq_active) return;
+
+	if (current_sample_rate > 0 && limiter_coeff_sample_rate != current_sample_rate) {
+		limiter_update_coefficients(current_sample_rate);
+	}
+
+	float peak = fmaxf(fabsf(*l), fabsf(*r));
+	float target_gain = 1.0f;
+	if (peak > LIMITER_THRESHOLD) {
+		target_gain = LIMITER_THRESHOLD / (peak + 1e-20f);
+	}
+
+	if (target_gain < limiter_gain) {
+		limiter_gain = target_gain + (limiter_attack_coeff * (limiter_gain - target_gain));
+	} else {
+		limiter_gain = target_gain + (limiter_release_coeff * (limiter_gain - target_gain));
+	}
+
+	if (!isfinite(limiter_gain) || limiter_gain <= 0.0f) limiter_gain = 1.0f;
+
+	*l *= limiter_gain;
+	*r *= limiter_gain;
+
+	// final guard against any possible hard clipping
+	if (*l > 1.0f) *l = 1.0f;
+	else if (*l < -1.0f) *l = -1.0f;
+	if (*r > 1.0f) *r = 1.0f;
+	else if (*r < -1.0f) *r = -1.0f;
+}
+
 int get_audio(int max, float* buff) {
 		int b = 0;
 
@@ -1397,6 +1634,9 @@ int get_audio(int max, float* buff) {
 		else if ((mode == PLAYING || mode == RAMP_DOWN || mode == ENDING) && get_buff_fill() > 0 && buffering == 0) {
 
 			//pthread_mutex_lock(&buffer_mutex);
+			if (eq_enabled && current_sample_rate > 0 && (eq_dirty || eq_coeff_sample_rate != current_sample_rate)) {
+				eq_rebuild_coefficients(current_sample_rate);
+			}
 
 			b = 0; // byte number
 
@@ -1455,6 +1695,7 @@ int get_audio(int max, float* buff) {
 
 				float l = bfl[low];
 				float r = bfr[low];
+				eq_process_stereo(&l, &r);
 
 				if (fabs(l) > peak_roll_l) peak_roll_l = fabs(l);
 				if (fabs(r) > peak_roll_r) peak_roll_r = fabs(r);
@@ -1468,8 +1709,10 @@ int get_audio(int max, float* buff) {
 
 				// Apply final volume adjustment
 				float final_vol = pow((gate * volume_on), config_volume_power);
+				if (eq_enabled && eq_active) final_vol *= eq_headroom_gain;
 				l = l * final_vol;
 				r = r * final_vol;
+				limiter_process_stereo(&l, &r);
 
 				buff[b] = l;
 				buff[b + 1] = r;
@@ -3250,6 +3493,43 @@ EXPORT int ramp_volume(int percent, int speed) {
 	return 0;
 }
 
+EXPORT void eq_set_enable(int n) {
+	pthread_mutex_lock(&buffer_mutex);
+	eq_enabled = (n != 0);
+	eq_dirty = true;
+	if (!eq_enabled) {
+		eq_headroom_db = 0.0f;
+		eq_headroom_gain = 1.0f;
+		eq_reset_state();
+	}
+	limiter_reset_state();
+	pthread_mutex_unlock(&buffer_mutex);
+}
+
+EXPORT void eq_set_band(int band, float gain_db) {
+	if (band < 0 || band >= EQ_BAND_COUNT) return;
+	if (gain_db > 12.0f) gain_db = 12.0f;
+	if (gain_db < -12.0f) gain_db = -12.0f;
+
+	pthread_mutex_lock(&buffer_mutex);
+	eq_band_gain_db[band] = gain_db;
+	eq_dirty = true;
+	pthread_mutex_unlock(&buffer_mutex);
+}
+
+EXPORT void eq_reset() {
+	pthread_mutex_lock(&buffer_mutex);
+	for (int i = 0; i < EQ_BAND_COUNT; i++) {
+		eq_band_gain_db[i] = 0.0f;
+	}
+	eq_dirty = true;
+	eq_headroom_db = 0.0f;
+	eq_headroom_gain = 1.0f;
+	eq_reset_state();
+	limiter_reset_state();
+	pthread_mutex_unlock(&buffer_mutex);
+}
+
 EXPORT int get_position_ms() {
 	if (command != START && command != LOAD && !reset_set && current_sample_rate > 0) {
 		return (int) ((position_count / (float) current_sample_rate) * 1000.0);
